@@ -21,11 +21,15 @@
  */
 
 
-#include "gmime-part.h"
-#include "gmime-utils.h"
 #include <config.h>
+#include <stdio.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
 #include <string.h>
 #include <ctype.h>
+#include "gmime-part.h"
+#include "gmime-utils.h"
 
 
 /**
@@ -122,7 +126,8 @@ g_mime_part_destroy (GMimePart *mime_part)
 		g_list_free (mime_part->children);
 	}
 	
-	g_free (mime_part->content);
+	if (mime_part->content)
+		g_byte_array_free (mime_part->content, TRUE);
 	
 	g_free (mime_part);
 }
@@ -467,19 +472,55 @@ g_mime_part_get_filename (GMimePart *mime_part)
 }
 
 
+static void
+read_random_pool (gchar *buffer, size_t bytes)
+{
+	int fd;
+	
+	fd = open ("/dev/urandom", O_RDONLY);
+	if (fd == -1) {
+		fd = open ("/dev/random", O_RDONLY);
+		if (fd == -1)
+			return;
+	}
+	
+	read (fd, buffer, bytes);
+	close (fd);
+}
+
+
 /**
  * g_mime_part_set_boundary: Set the multi-part boundary (not used on non-multiparts)
  * @mime_part: Mime part
- * @boundary: the boundary for the multi-part
+ * @boundary: the boundary for the multi-part or NULL to generate a random one.
  *
  * Sets the boundary on the mime part.
  **/
 void
 g_mime_part_set_boundary (GMimePart *mime_part, const gchar *boundary)
 {
+	gchar bbuf[27];
+	
 	g_return_if_fail (mime_part != NULL);
 	
 	g_free (mime_part->boundary);
+	
+	if (!boundary) {
+		/* Generate a fairly random boundary string. */
+		char digest[16], *p;
+		int state, save;
+		
+		read_random_pool (digest, 16);
+		
+		strcpy (bbuf, "=-");
+		p = bbuf + 2;
+		state = save = 0;
+		p += g_mime_utils_base64_encode_step (digest, 16, p, &state, &save);
+		*p = '\0';
+		
+		boundary = bbuf;
+	}
+	
 	mime_part->boundary = g_strdup (boundary);
 }
 
@@ -502,17 +543,86 @@ g_mime_part_get_boundary (GMimePart *mime_part)
 /**
  * g_mime_part_set_content: Set the content of the mime part
  * @mime_part: Mime part
- * @content: Mime part content
+ * @content: raw mime part content
+ * @len: raw content length
  *
  * Sets the content of the Mime Part (only non-multiparts)
  **/
 void
-g_mime_part_set_content (GMimePart *mime_part, const char *content)
+g_mime_part_set_content (GMimePart *mime_part, const char *content, guint len)
 {
 	g_return_if_fail (mime_part != NULL);
 	
-	g_free (mime_part->content);
-	mime_part->content = g_strdup (content);
+	if (mime_part->content)
+		g_byte_array_free (mime_part->content, TRUE);
+	
+	mime_part->content = g_byte_array_new ();
+	g_byte_array_append (mime_part->content, content, len);
+}
+
+
+/**
+ * g_mime_part_set_pre_encoded_content: Set the pre-encoded content of the mime part
+ * @mime_part: Mime part
+ * @content: encoded mime part content
+ * @len: length of the content
+ * @encoding: content encoding
+ *
+ * Sets the encoding type and raw content on the mime part after decoding the content.
+ **/
+void
+g_mime_part_set_pre_encoded_content (GMimePart *mime_part, const char *content, guint len, GMimePartEncodingType encoding)
+{
+	gchar *raw;
+	gint save = 0, state = 0;
+	
+	g_return_if_fail (mime_part != NULL);
+	g_return_if_fail (content != NULL);
+	
+	if (mime_part->content)
+		g_byte_array_free (mime_part->content, TRUE);
+	
+	mime_part->content = g_byte_array_new ();
+	g_byte_array_set_size (mime_part->content, len);
+	raw = mime_part->content->data;
+	switch (encoding) {
+	case GMIME_PART_ENCODING_BASE64:
+		len = g_mime_utils_base64_decode_step (content, len, raw, &state, &save);
+		g_byte_array_set_size (mime_part->content, len);
+		break;
+	case GMIME_PART_ENCODING_QUOTEDPRINTABLE:
+		len = g_mime_utils_quoted_decode_step (content, len, raw, &state, &save);
+		g_byte_array_set_size (mime_part->content, len);
+		break;
+	default:
+		memcpy (raw, content, len);
+		
+		/* do some smart 8bit detection */
+		if (encoding == GMIME_PART_ENCODING_DEFAULT && g_mime_utils_text_is_8bit (raw))
+			encoding = GMIME_PART_ENCODING_8BIT;
+	}
+	
+	mime_part->encoding = encoding;
+}
+
+
+/**
+ * g_mime_part_get_content: 
+ * @mime_part: the GMimePart to be decoded.
+ * @len: decoded length (to be set after processing)
+ * 
+ * Returns a gchar * pointer to the raw contents of the MIME Part
+ * and sets %len to the length of the buffer.
+ **/
+const gchar *
+g_mime_part_get_content (GMimePart *mime_part, guint *len)
+{
+	g_return_val_if_fail (mime_part != NULL, NULL);
+	g_return_val_if_fail (mime_part->content != NULL, NULL);
+	
+	*len = mime_part->content->len;
+	
+	return mime_part->content->data;
 }
 
 
@@ -544,15 +654,15 @@ get_content_disposition (GMimePart *mime_part)
 	
 	params = mime_part->disposition->params;
 	
-	if (mime_part->disposition->disposition && *mime_part->disposition->disposition) {
+	if (mime_part->disposition->disposition && *mime_part->disposition->disposition)
 		string = g_string_new (mime_part->disposition->disposition);
-		if (params)
-			g_string_append (string, ";");
-	} else {
+	else
 		string = g_string_new ("");
-	}
 	
-	while (TRUE) {
+	if (params)
+		g_string_append (string, ";");
+	
+	while (params) {
 		GMimeParam *param;
 		gchar *buf;
 		
@@ -617,6 +727,35 @@ get_content_type (GMimeContentType *mime_type)
 	return str;
 }
 
+static gchar *
+get_content (GMimePart *part)
+{
+	gchar *content;
+	gint save = 0, state = 0;
+	gint len;
+	
+	if (!part->content)
+		return g_strdup ("");
+	
+	switch (part->encoding) {
+	case GMIME_PART_ENCODING_BASE64:
+		content = g_malloc (BASE64_ENCODE_LEN (part->content->len));
+		len = g_mime_utils_base64_encode_close (part->content->data, part->content->len, content, &state, &save);
+		content[len] = '\0';
+		break;
+	case GMIME_PART_ENCODING_QUOTEDPRINTABLE:
+		state = -1;
+		content = g_malloc (QP_ENCODE_LEN (part->content->len));
+		len = g_mime_utils_quoted_encode_close (part->content->data, part->content->len, content, &state, &save);
+		content[len] = '\0';
+		break;
+	default:
+		content = g_strndup (part->content->data, part->content->len);
+	}
+	
+	return content;
+}
+
 /**
  * g_mime_part_to_string: Write the MIME Part to a string
  * @mime_part: MIME Part
@@ -637,6 +776,10 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel) {
 		GList *child;
 		gchar *content_type;
 		
+		/* make sure there's a boundary, else force a random boundary */
+		if (!mime_part->boundary)
+			g_mime_part_set_boundary (mime_part, NULL);
+		
 		content_type = g_mime_content_type_to_string (mime_part->mime_type);
 		
 		contents = g_string_new ("");
@@ -679,6 +822,7 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel) {
 		gchar *disposition;
 		gchar *description;
 		gchar *content_id;
+		gchar *content;
 		gchar *extras;
 		gchar *text;
 		
@@ -703,6 +847,8 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel) {
 		else
 			extras = "";
 		
+		content = get_content (mime_part);
+		
 		if (disposition && *disposition) {
 			string = g_strdup_printf ("%sContent-Type: %s\n"
 						  "Content-Transfer-Encoding: %s\n"
@@ -715,7 +861,7 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel) {
 						  disposition,
 						  description,
 						  content_id,
-						  mime_part->content);
+						  content);
 		} else {
 			string = g_strdup_printf ("%sContent-Type: %s\n"
 						  "Content-Transfer-Encoding: %s\n"
@@ -726,13 +872,14 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel) {
 						  g_mime_part_encoding_to_string (mime_part->encoding),
 						  description,
 						  content_id,
-						  mime_part->content);
+						  content);
 		}
 		
 		g_free (content_type);
 		g_free (disposition);
 		g_free (description);
 		g_free (content_id);
+		g_free (content);
 	}
 	
 	return string;
@@ -740,37 +887,28 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel) {
 
 
 /**
- * g_mime_part_decode_contents: Convenience MIME Part decoding function.
- * @mime_part: the GMimePart to be decoded.
- * @len: decoded length (to be set after processing)
+ * g_mime_part_foreach: 
+ * @mime_part: the MIME part
+ * @callback: function to call for #mime_part and all it's subparts
+ * @data: extra data to pass to the callback
  * 
- * Returns a gchar * pointer to the decoded contents of the MIME Part
- * and sets %len to the length of the decoded buffer.
+ * Calls #callback on #mime_part and each of it's subparts.
  **/
-gchar *
-g_mime_part_decode_contents (GMimePart *mime_part, guint *len)
+void
+g_mime_part_foreach (GMimePart *mime_part, GMimePartFunc callback, gpointer data)
 {
-	GMimePartEncodingType encoding;
-	const gchar *content;
-	gchar *body;
-	int state = 0, save = 0;
+	g_return_if_fail (mime_part != NULL);
+	g_return_if_fail (callback != NULL);
 	
-	encoding = g_mime_part_get_encoding (mime_part);
-	content = mime_part->content;
-	*len = strlen (content);
+	callback (mime_part, data);
 	
-	switch (encoding) {
-	case GMIME_PART_ENCODING_BASE64:
-		body = g_malloc (*len);
-		*len = g_mime_utils_base64_decode_step (content, *len, body, &state, &save);
-		break;
-	case GMIME_PART_ENCODING_QUOTEDPRINTABLE:
-		body = g_malloc (*len);
-		*len = g_mime_utils_quoted_decode_step (content, *len, body, &state, &save);
-		break;
-	default:
-		body = g_strdup (content);
+	if (mime_part->children) {
+		GList *child;
+		
+		child = mime_part->children;
+		while (child) {
+			g_mime_part_foreach ((GMimePart *) child->data, callback, data);
+			child = child->next;
+		}
 	}
-	
-	return body;
 }
